/*************************************************************
*  This file is part of the Surface Evolver source code.     *
*  Programmer:  Ken Brakke, brakke@susqu.edu                 *
*************************************************************/

/*************************************************************
*
*     file:    utility.c
*
*     Purpose:    Various utility routines.
*/

#include "include.h"

/**************************************************************
*  
*  Function: catcher()
*
*  Purpose:  User interrupt catcher.  Sets breakflag so 
*        anybody in a long loop can test to see if
*        user wants to break out.
*/

#ifdef __WIN32__
#define write(a,b,c) erroutstring(b)
#endif

void catcher(sig)
int sig;
{
  if ( sig == SIGINT ) 
     { signal(SIGINT,catcher);

#ifdef SHARED_MEMORY
    if ( mpflag == M_ACTIVE )
      { if ( m_breakflag[m_get_myid()] )
        longjmp(m_jumpbuf[m_get_myid()],1);
         else m_breakflag[m_get_myid()] = 1;
         if ( m_get_myid() == 0 ) 
        write(2,"\nWill break after iteration.\n",29);
         breakflag = 1;
         iterate_flag = 0;
      }
    else
      if ( m_get_myid() > 0 ) {}
    else
#endif
    if ( iterate_flag == 1 )
      { write(2,"\nWill break after iteration.\n",29);
         breakflag = 1;
         iterate_flag = 0;
      }
    else if ( iterate_flag == 2 ) /* graphing or something */
      { write(2,"\nWill abort operation.\n",24);
         breakflag = 1;
         iterate_flag = 0;
      }
    else
      { write(2,"\nAborting operation.\n",22);
#ifdef WIN32
/* separate thread for signal handler, so can't longjmp out of error */
  erroutstring("Can't nicely abort command in WIN32.  Hang in there, or\n");
  erroutstring("use Task Manager to kill Evolver if you think it's hung.\n");
#else
         kb_error(1357,"",RECOVERABLE_QUIET);
#endif
      }
    while ( commandfd && (commandfd  != stdin) )
         pop_commandfd();
    quiet_flag = 0; 
   }
    
#ifdef SIGTERM
  if ( sig == SIGTERM )  /* for outside kill and dump */
    { sprintf(errmsg,"Caught SIGTERM. proc %d ",getpid());
      erroutstring(errmsg);
      do_dump(NULL); /* dump to default */
      my_exit(1);
    }
#endif

#ifdef SIGHUP
  if ( sig == SIGHUP )  /* for outside kill and dump */
    {
      do_dump(NULL); /* dump to default */
      my_exit(1);
    }
#endif

#ifdef SIGPIPE
  else if ( sig == SIGPIPE )
     { signal(SIGPIPE,catcher);
    write(2,"Broken pipe signal.\n",20);
    broken_pipe_flag = 1;
     }
#endif
}

/***********************************************************************
*
* function: kb_checksum()
*
* purpose: to calculate a simple checksum on a block of memory.
*/

int kb_checksum(mem,size)
char *mem;
int    size;
{
  int sum = 0;
  int *a,*b;
  b = (int *)mem;
  a = b + size/sizeof(int);
  do { sum += *a; } while ( --a != b);
  return sum;
}

/*********************************************************************
*
*  function: kb_calloc
*
*  purpose: memory allocation with error detection
*          Also memory debugging.
*          Called by macro mycalloc(num,size).
*/

#ifdef MEMSTRINGS
char *kb_calloc(num,size,file,line)
int num,size;
char *file;
int line;
#else
char *kb_calloc(num,size)
int num,size;
#endif
{ char *ptr;
  
  if ( (num < 0) || (size <0) )
     kb_error(1358,"Internal error: Trying to allocate negative size memory block.\n",RECOVERABLE);

  if ( num == 0 ) num = 1; /* to prevent problem with zero block size */
  if ( size == 0 ) size = 1;
  ptr = (char *)calloc(num,size);
  if ( ptr == NULL )
     kb_error(1359,"Internal error: Cannot allocate memory.\n",RECOVERABLE);

if ( memdebug )
{
#ifdef MEMSTRINGS
    sprintf(errmsg,"%11.11s %4d: ",file,line);
    outstring(errmsg);
#endif

#if defined(M_MXFAST) && defined(IRIS)
  {  struct mallinfo m;
    m = mallinfo();
    sprintf(errmsg,"mycalloc %X  %d*%d = %d    Arena %d  orduse %d\n",
        ptr,num,size,num*size, m.arena,m.uordblks);
    outstring(errmsg); 
  }
#else
  sprintf(errmsg,"mycalloc %X  %d*%d = %d \n",ptr,num,size,num*size);
  outstring(errmsg); 
#endif
}

  return ptr;
}  


/*********************************************************************
*
*  function: KB_realloc
*
*  purpose: memory re-allocation with error detection
*          Needs oldsize as input so can initialize new memory to 0
*          Called through macro kb_realloc for optional memory debug.
*/

#ifdef MEMSTRINGS
char *KB_realloc(ptr,size,oldsize,file,line)
char *ptr;
int size;
int oldsize;
char *file; int line;
#else
char *KB_realloc(ptr,size,oldsize)
char *ptr;
int size;
int oldsize;
#endif
{ char *newptr;

  if ( ptr == NULL ) return mycalloc(size,1);
  if ( size <= 0 ) size = 1;
  newptr = (char *)realloc(ptr,size);
  if ( newptr == NULL )
  { sprintf(errmsg,"Internal error: Cannot reallocate memory from %d to %d\n",oldsize,size);
     kb_error(1360,errmsg,RECOVERABLE);
  }
  else if (oldsize < size )  memset(newptr+oldsize,0,size-oldsize);
  if ( memdebug )
  { 
#ifdef MEMSTRINGS
     sprintf(errmsg,"%11.11s %4d: recalloc old %X new %X size %d\n",
         file,line,ptr,newptr,size);
#else
     sprintf(errmsg,"recalloc old %X new %X size %d\n",ptr,newptr,size);
#endif
     outstring(errmsg);
  }
  return newptr;
}  


/********************************************************************
*
* function:  temp_calloc()
*
* purpose:  memory allocation that can be undone by longjmp at error.
*
*/

#define MAXMEMS 200
static char *memlist[MAXMEMS];  /* allocated blocks, packed list */
static int    memcount;         /* number allocated */

char *temp_calloc(num,size)
int num,size;
{
  char *ptr;

  if ( memcount > (MAXMEMS - 1) )
     kb_error(1361,"Internal error: Too many temporary memory blocks.",RECOVERABLE);

  ptr = mycalloc(num,size);
  memlist[memcount++] = ptr;
  return ptr;
}  


/********************************************************************
*
* function:  temp_realloc()
*
* purpose:  Changing size of temporary memory allocation.
*          Needs oldsize as input so can initialize new memory to 0
*
*/

char * temp_realloc(ptr,size,oldsize)
char *ptr;
int size;
int oldsize;
{
  int k;

  for ( k = 0 ; k < memcount ; k++ )
     if ( ptr == memlist[k] ) break;
  if ( k == memcount )
  { kb_error(1362,"Internal error: Cannot find address in temporary memory list.",WARNING);
    return NULL;
  }
  else
     memlist[k] = ptr = kb_realloc(ptr,size,oldsize);
  return ptr;
}


/********************************************************************
*
* function:  temp_free()
*
* purpose:  Usual freeing of temporary memory allocation.
*
*/

void temp_free(ptr)
char *ptr;
{
  int k;

  for ( k = 0 ; k < memcount ; k++ )
     if ( ptr == memlist[k] ) break;
  if ( k == memcount )
  { 
    /*kb_error(1363,"Internal error: Cannot find address in temporary memory list.",WARNING);
      */
    return;
  }
  else
  { myfree(ptr);
    memlist[k] = memlist[--memcount];  /* pack list */
    memlist[memcount] = NULL;
  }
}

/********************************************************************
*
* function:  temp_free_all()
*
* purpose:  Freeing of all temporary memory in case of error.
*
*/


void temp_free_all()
{
  int k;

#ifdef MSC
  while ( graph_semaphore ) ;
#endif

  for ( k = 0 ; k < memcount ; k++ )
  { myfree(memlist[k]);
    memlist[k] = NULL;
  }
  memcount = 0;

  /* pointers to be annulled */
  saved.coord = NULL;
  vhead = NULL;
  array = NULL;
  pressures = NULL;
  congrads = NULL;
  conrhs = NULL;

}

/********************************************************************
*
* function:  myfree()
*
* purpose:  Freeing memory.  Common call for all memory freeing.
*
*/

void myfree(ptr)
char *ptr;
{
  if ( memdebug ) { sprintf(errmsg,"Freeing %X\n",ptr); 
  outstring(errmsg); }

  if ( ptr ) free(ptr);

}



/********************************************************************

    Memory allocation routines for the block of memory holding
    dynamic structures of surface description.  Enables
    exporting surface info to multiple processors.

    Returns are offsets from start of dynamic region.

*********************************************************************/
 
#define ALIGNSIZE 8
struct dy_head { int size; /* bytes in block, including structures */
              int state; /* see defines; also used as magic number */
              };
#define DY_FREE  0x2913
#define DY_USED  0x9177
struct dy_end { int size; int pad; };
#define DY_STARTSIZE 0x1000

/* for doubly linked list of free blocks */
struct dy_free { DY_OFFSET prev;
              DY_OFFSET next;
              };
int dy_minsize = sizeof(struct dy_free);
int dy_minfrag = 2*sizeof(struct dy_head) + sizeof(struct dy_free);
int dy_overhead = sizeof(struct dy_head) + sizeof(struct dy_end);


DY_OFFSET dy_calloc(num,size)
int num;
int size;
{ struct dy_head *ptr,*newptr;
  struct dy_free *freeptr,*newfptr;
  struct dy_end  *endptr;
  DY_OFFSET off;
  int needed;
  int surplus;
  int asize;

  if ( dymemsize == 0 )  /* initial allocation */
  { if ( sizeof(struct dy_head) % ALIGNSIZE )
      { sprintf(errmsg,"Internal error: sizeof(struct dy_head) %d not multiple of ALIGNSIZE %d\n",
        sizeof(struct dy_head),ALIGNSIZE);
         kb_error(1364,errmsg,UNRECOVERABLE);

      }
    if ( sizeof(struct dy_end) % ALIGNSIZE )
      { sprintf(errmsg,"Internal error: ALIGNSIZE %d != sizeof(struct dy_end) %d\n",
        ALIGNSIZE,sizeof(struct dy_end));
         kb_error(1365,errmsg,UNRECOVERABLE);

      }
    dymem = mycalloc(DY_STARTSIZE,1);
    dymemsize = DY_STARTSIZE;
    ptr = (struct dy_head *)dymem;
    ptr->size = DY_STARTSIZE;
    ptr->state = DY_FREE;

    dy_freestart = sizeof(struct dy_head);
    freeptr = (struct dy_free *)(ptr + 1);
    freeptr->prev = 0; /* first in list */
    freeptr->next = 0; /* last in list */

    endptr = (struct dy_end *)(dymem + DY_STARTSIZE) - 1;
    endptr->size = DY_STARTSIZE;
  }

  /* size of needed block */
  asize = ((num*size + ALIGNSIZE - 1)/ALIGNSIZE)*ALIGNSIZE;
  if ( asize <= dy_minsize ) asize = dy_minsize;
  needed = sizeof(struct dy_head) + asize + sizeof(struct dy_end);

  if ( memdebug )
  { sprintf(errmsg,"dy_alloc %d*%d = %d bytes.\n",num,size,asize);
     outstring(errmsg);
  }

retry_alloc:

  /* look for first fit in free block list */
  if ( dy_freestart )
    for ( off = dy_freestart ; off > 0 ; off = freeptr->next )
    { freeptr = (struct dy_free *)(dymem + off);
      ptr = (struct dy_head *)freeptr - 1;
      /* check size */
      if ( needed > ptr->size )
      { if ( freeptr->next == 0 ) break;
        else continue;
      }
      surplus = ptr->size - needed;
      if ( surplus > dy_minfrag )
      { /* divide block */
        newptr = (struct dy_head *)(dymem + off + needed) - 1;
        newptr->size = ptr->size - needed;
        newptr->state = DY_FREE;
        newfptr = (struct dy_free *)(newptr + 1);
        *newfptr = *freeptr;
        endptr = (struct dy_end*)((char*)ptr + ptr->size) - 1;
        endptr->size -= needed;
        ptr->size = needed;
        endptr = (struct dy_end*)((char*)ptr + ptr->size) - 1;
        endptr->size = needed;
        if ( freeptr->prev )
         ((struct dy_free *)(dymem + freeptr->prev))->next = off + needed;
        else dy_freestart = off + needed;
        if ( freeptr->next )
         ((struct dy_free *)(dymem + freeptr->next))->prev = off + needed;
      }
      else
      { /* remove from freelist */
        if ( freeptr->prev )
           ((struct dy_free *)(dymem + freeptr->prev))->next = freeptr->next;
        else dy_freestart = freeptr->next;
        if ( freeptr->next )
           ((struct dy_free *)(dymem + freeptr->next))->prev = freeptr->prev;
      }
      memset((char*)freeptr,0,sizeof(struct dy_free));
      ptr->state = DY_USED;

      /* some handy debugging equivalences */
      Gen_quants = GEN_QUANTS;
      Meth_inst = METH_INST;
      Globals = globals;

      return off;
    }

  /* if here, then no block big enough */
  if ( memdebug ) outstring ("expanding dymem.\n");
  dymem = kb_realloc(dymem,2*dymemsize,dymemsize);
  endptr = (struct dy_end *)(dymem + dymemsize) - 1; /* old end */
  ptr = (struct dy_head *)(dymem + dymemsize - endptr->size);

  if ( ptr->state == DY_FREE )
    { /* combine with last block */
      ptr->size += dymemsize;
      memset((char*)endptr,0,sizeof(struct dy_end));
      endptr = (struct dy_end *)(dymem + 2*dymemsize) -1;
      endptr->size = ptr->size;
    }
  else 
    { /* put in free block list at head */
      endptr = (struct dy_end *)(dymem + 2*dymemsize) - 1;
      ptr = (struct dy_head *)(dymem + dymemsize);
      freeptr = (struct dy_free *)(ptr + 1);
      ptr->size = dymemsize;
      ptr->state = DY_FREE;
      endptr->size = dymemsize;
      freeptr->prev = 0;
      if ( dy_freestart )
         ((struct dy_free *)(dymem+dy_freestart))->prev = 
             (char*)freeptr - dymem;
      dy_freestart = (char*)freeptr - dymem;
    }
      
  dymemsize = 2*dymemsize;
  goto retry_alloc;
}

DY_OFFSET dy_realloc(old,newnum,newsize)
DY_OFFSET old;
int newnum;
int newsize;
{ DY_OFFSET new;
  struct dy_head *ptr;

  if ( memdebug )
  { outstring("dy_realloc "); }

  new = dy_calloc(newnum,newsize);
  if ( old )
  { ptr = (struct dy_head *)(dymem + old) - 1;
    if ( ptr->size < newnum*newsize )
      memcpy(dymem+new,dymem+old,ptr->size);
    else  memcpy(dymem+new,dymem+old,newnum*newsize);
    dy_free(old);
  }

  /* some handy debugging equivalences */
  Gen_quants = GEN_QUANTS;
  Meth_inst = METH_INST;
  Globals = globals;

  return new;
}

void dy_free(spot)
DY_OFFSET spot;
{ struct dy_head *ptr = (struct dy_head *)(dymem + spot) - 1;
  struct dy_free *freeptr = (struct dy_free *)(dymem + spot);

  if ( ptr->state != DY_USED )
     kb_error(1366,"Internal error: Trying to dy_free unallocated block.\n",RECOVERABLE);

  memset(dymem+spot,0,ptr->size - dy_overhead);
  freeptr->next = dy_freestart;
  freeptr->prev = 0;
  if ( dy_freestart )
     ((struct dy_free *)(dymem + dy_freestart))->prev = spot;
  dy_freestart = spot;
  ptr->state = DY_FREE;
  
  /* see if can combine with nbrs */
    
}

/********* end of dynamic memory stuff **************************/


/**************************************************************
*
*  Function: calc_edge()
*
*  Purpose:  Calculate the length of an edge from its endpoint
*        coordinates.  Puts result into edge structure.
*
*/

void calc_edge(e_id)
edge_id e_id;
{
  REAL len;
  REAL s[MAXCOORD];
  REAL midx[MAXCOORD];
  int i,j,k;
  vertex_id tv = get_edge_tailv(e_id);
  REAL *xt=get_coord(tv);
  REAL euclidean;

  get_edge_side(e_id,s);
  euclidean = SDIM_dot(s,s);
  if ( klein_metric_flag )
     { for ( i = 0 ; i < SDIM ; i++ ) midx[i] = xt[i] + s[i];/*unwrap head*/
    len = klein_length(xt,midx);
     }
  else if ( web.metric_flag )
  {
     /* energy due to linear tension, metric evaluated at midpoint */
     for ( k = 0, len = 0.0 ; k < gauss1D_num ; k++ )
     { REAL sum;
       for ( i = 0 ; i < SDIM ; i++ )
          midx[i] = gauss1Dpt[k]*s[i] + xt[i];
       if ( web.conformal_flag )
       { REAL gg = eval(&web.metric[0][0],midx,NULLID);
         len += gauss1Dwt[k]*sqrt(gg*euclidean);
       }
       else
       { for ( sum = 0.0, i = 0 ; i < SDIM ; i++ )
         for ( j = 0 ; j < SDIM ; j++ )
          sum += s[i]*s[j]*eval(&web.metric[i][j],midx,NULLID);
         len += gauss1Dwt[k]*sqrt(sum);
       }
     }
  }
  else
  {
    switch (web.modeltype)
    { case LINEAR: len = sqrt(euclidean); break;
      case QUADRATIC:
      { MAT2D(x,3,MAXCOORD);
        REAL tang[MAXCOORD];
        int m;

        len = 0;
        get_edge_verts(e_id,x,NULL);
        for ( m = 0 ; m < gauss1D_num ; m++ )
        { for ( j = 0 ; j < SDIM ; j ++ )
         { tang[j] = 0.0;
            for ( k = 0 ; k < edge_ctrl ; k++ )
            tang[j] += gauss1polyd[k][m]*x[k][j];
         }
          len += gauss1Dwt[m]*sqrt(SDIM_dot(tang,tang));
        }
     } break;
    case LAGRANGE:
    { /* tangent vectors at gauss points */
      int m;
      MAT2D(x,20,MAXCOORD);
      REAL tang[MAXCOORD];
      int ctrl = web.skel[EDGE].ctrlpts;
      struct gauss_lag *gl = &gauss_lagrange[1][web.gauss1D_order];
      len = 0.0;
      get_edge_verts(e_id,x,NULL);
      for ( m = 0 ; m < gl->gnumpts ; m++ )
      {  vec_mat_mul(gl->gpolypart[m][0],x,tang,ctrl,SDIM);
          len += gl->gausswt[m]*sqrt(SDIM_dot(tang,tang));
      }

    }

    break; 

   } /* end switch */
 }
 set_edge_length(e_id,len);
}

/************************************************************
*
*  get_edge_side()
*
*  Purpose:  calculate the vector of an edge
*
*  Input:     edge_id e_id - specifies edge
*        REAL  *x;  - 3 components of vector
*
*  Output:    x[] has side components
*/

void get_edge_side(e_id,x)
edge_id e_id;
REAL *x;
{
  REAL *y,*z;
  int i;
  REAL w[MAXCOORD];
  
  y = get_coord(get_edge_tailv(e_id));
  z = get_coord(get_edge_headv(e_id));
  if ( web.symmetry_flag ) 
  {
    (*sym_wrap)(z,w,get_edge_wrap(e_id));
    z = w;
  }
  for ( i = 0 ; i < SDIM ; i++ ) 
      x[i] = z[i] - y[i];
}

/*************************************************************
*
*  function: get_edge_adjust()
*
*  purpose:  adjust a side vector for torus wrap.
*        used only in torvol.c.
*
*/
  
void get_edge_adjust(e_id,w)
edge_id e_id;
REAL *w;
{
  REAL x[MAXCOORD];
  int i ;

  for ( i = 0 ; i < SDIM ; i++ ) x[i] = 0.0;
  (*sym_wrap)(x,w,get_edge_wrap(e_id));
}
             


/***************************************************************
*
*  Function:  get_vertex_evalence
*
*  Purpose:  return number of edges around vertex
*
*/

int get_vertex_evalence(v_id)
vertex_id v_id;
{
  int n = 0;
  edge_id e_id;
  edge_id next_e;
  int attr = get_vattr(v_id);

  if ( attr & (Q_MIDPOINT|Q_MIDEDGE) ) return 1;
  if ( vedge_timestamp < top_timestamp ) make_vedge_lists();
  e_id = get_vertex_edge(v_id);
  if ( !valid_element(e_id) ) return 0;
  next_e = e_id;
  do
  { n++;
    next_e = get_next_tail_edge(next_e);
  }
  while ( !equal_element(next_e,e_id) && (n < web.skel[EDGE].count) );
  return n;
}


/***************************************************************
*
*  Function:  get_edge_valence
*
*  Purpose:  return number of facets around edge
*
*/

int get_edge_valence(e_id)
edge_id e_id;
{
  int n = 0;
  facetedge_id fe = get_edge_fe(e_id);
  facetedge_id next_fe = fe;

  if ( !valid_id(fe) ) return 0;
  do
  { if ( valid_id(get_fe_facet(fe)) ) n++;
    next_fe = get_next_facet(next_fe);
  }
  while ( next_fe != fe );
  return n;
}



/***************************************************************
*
*  Function:  get_facet_valence
*
*  Purpose:  return number of edges around facet
*
*/

int get_facet_valence(f_id)
facet_id f_id;
{
  int n = 0;
  facetedge_id fe = get_facet_fe(f_id);
  facetedge_id next_fe = fe;

  if ( !valid_id(fe) ) return 0;
  do
  { n++;
    next_fe = get_next_edge(next_fe);
  }
  while ( valid_id(next_fe) && (next_fe != fe) );
  return n;
}

/***************************************************************
*
*  Function:  get_body_valence
*
*  Purpose:  return number of facets around a body
*
*/

int get_body_valence(b_id)
body_id b_id;
{
  int n = 0;
  facet_id f_id = get_fe_facet(get_body_fe(b_id));
  facet_id next_f = f_id;

  if ( !valid_id(f_id) ) return 0;
  if ( web.representation == STRING ) return get_facet_valence(f_id);
  if ( bfacet_timestamp < top_timestamp )
  { make_bfacet_lists();
    bfacet_timestamp = top_timestamp;
  }
  do
  { n++;
    next_f = get_next_body_facet(next_f);
  }
  while ( (next_f != f_id) && (n < 3*web.skel[FACET].count) );
  return n;
}

/*********************************************************************
*
* function: set_facet_vertices()
*
* purpose: Fill in vertex lists of facets.
*/
void set_facet_vertices()
{
  /* have to make expandable v list for quadratic mode */
}
/****************************************************************
*
*  function: begin_normal_motion()
*
*  purpose: calculates vertex normals as volume gradients
*
*/

void begin_normal_motion()
{
  vertex_id v_id;
  facetedge_id fe;
  
  if ( vertex_normals != NULL ) myfree((char*)vertex_normals);
  vertex_normals = (pt_type *)mycalloc(
     web.skel[VERTEX].max_ord+1,sizeof(pt_type));

  FOR_ALL_VERTICES(v_id)
    { fe = get_vertex_fe(v_id);
      if ( !valid_id(fe) ) continue;
      calc_vertex_normal(v_id,fe,vertex_normals[ordinal(v_id)]);
    }
  normal_motion_flag = 1;
}

/*****************************************************************
*
* function: end_normal_motion()
*
* purpose: free memory used in normal motion mode
*
*/

void end_normal_motion()
{
  if ( vertex_normals != NULL ) myfree((char*)vertex_normals);
  vertex_normals = NULL;
  normal_motion_flag = 0;
}



/***********************************************************
*
*  Function: calc_vertex_normal
*
*  Purpose: Calculate average normal of facets around a vertex
*          for Pixar normal interpolation.  Only does the facets
*          in the same face as given facet; will not cross
*          multiple edges.
*
*  Input:    vertex id
*          facet-edge id - with tail at vertex, for defining face
*          norm -  pointer to place for normal vector
*
*
*  Output:  Average normal, normalized to unit length.
*
*  Return value: Length of original normal, i.e. volume gradient.
*/

REAL calc_vertex_normal(v_id,fe,norm)
vertex_id v_id;
facetedge_id fe;
REAL *norm;
{
  int i,j;
  REAL y[MAXCOORD],z[MAXCOORD];
  REAL fnorm[MAXCOORD];
  facetedge_id sidea;
  struct boundary *bdry;  /* to check staying on same boundary */
  REAL size;


  /* check for string */
  if ( web.representation == STRING )
  { edge_id e_id,ee_id;
    if ( vedge_timestamp < top_timestamp ) make_vedge_lists();
    e_id = get_vertex_edge(v_id);
    ee_id = get_next_tail_edge(e_id);

    get_edge_side(e_id,z);
    get_edge_side(ee_id,y);
    if ( equal_id(e_id,ee_id) ) { norm[0] = y[1]; norm[1] = -y[0]; }
    else { norm[0] = y[1] - z[1]; norm[1] = z[0] - y[0]; }
    if (inverted(e_id)) { norm[0] = -norm[0]; norm[1] = -norm[1];}
    goto normalize;
  }

  if ( !valid_id(fe) )
     kb_error(2001,"Internal error: Invalid fe in calc_vertex_normal().\n",RECOVERABLE);

  for ( i = 0 ; i < SDIM ; i++ ) norm[i] = 0.0;
  bdry = get_facet_boundary(get_fe_facet(fe));  /* original boundary */

  /* go around one way to edge of face */
  sidea = fe;
  do
  {
    get_fe_side(sidea,z);
    get_fe_side(get_prev_edge(sidea),y);
    cross_prod(y,z,fnorm);
    for ( i = 0 ; i < SDIM ; i++ ) norm[i] += fnorm[i];

    /* go to next facet */
    sidea = get_prev_edge(sidea);
    if ( equal_id(sidea,get_next_facet(sidea)) ) break;  /* edge */
    sidea = fe_inverse(get_next_facet(sidea));
    if ( bdry != get_facet_boundary(get_fe_facet(sidea)) ) break;
    if ( equal_id(sidea,fe) ) goto onormalize; /* went all the way around */
  }
  while ( equal_id(get_next_facet(sidea),get_prev_facet(sidea)) );

  /* go around the other way */
  sidea = fe;
  while ( equal_id(get_next_facet(sidea),get_prev_facet(sidea)) )
  {
    if ( equal_id(sidea,get_next_facet(sidea)) ) break;  /* edge */
    sidea = fe_inverse(get_next_facet(sidea));
    sidea = get_next_edge(sidea);
    if ( bdry != get_facet_boundary(get_fe_facet(sidea)) ) break;

    get_fe_side(sidea,z);
    get_fe_side(get_prev_edge(sidea),y);
    cross_prod(y,z,fnorm);
    for ( i = 0 ; i < SDIM ; i++ ) norm[i] += fnorm[i];
  }

  /* orient */
  onormalize:
  if ( inverted(get_fe_facet(fe)) )
      for ( i = 0 ; i < SDIM ; i++ ) norm[i] = -norm[i];

  normalize:
  if ( get_vattr(v_id) & CONSTRAINT )
    {
      conmap_t * conmap = get_v_constraint_map(v_id);
      int oncount = 0,hitcount;
      struct constraint *con[MAXCONPER];
      int conlist[MAXCONPER];
      REAL perp[MAXCOORD];

      for ( j = 1 ; j <= (int)conmap[0] ; j++ )
        { 
          if ( conmap[j] & CON_HIT_BIT )
          { conlist[oncount] = conmap[j];
            con[oncount++] = get_constraint(conmap[j]);
          }
        }

      hitcount = constr_proj(TANGPROJ,oncount,con,get_coord(v_id),
                 norm,perp,conlist,DETECT,v_id);
      if ( hitcount != oncount )
        { clear_v_constraint_status(v_id);
          for ( j = 0 ; j < hitcount ; j++ )
             set_v_constraint_status(v_id,conlist[j]);
        }
      for ( j = 0 ; j < SDIM ; j++ )
        norm[j] -= perp[j];
     }
      
  size = sqrt(SDIM_dot(norm,norm));
  if ( size <= 0.0 ) return 0.0; 
  for ( i = 0 ; i < SDIM ; i++ ) norm[i] /= size;
  return size/web.simplex_factorial;
}

/***********************************************************
*
*  Function: calc_vertex_smooth_normal
*
*  Purpose: Calculate average normal of facets around a vertex
*          for grapics normal interpolation.  Only does the facets
*          in the same face as given facet; will not cross
*          multiple edges, or boundary or constraint edges
*          when different from facet.  Only uses first 3 dimensions.
*          Does not project normal to constraints.
*
*  Input:    vertex id
*          facet-edge id - with tail at vertex, for defining face
*          norm -  pointer to place for normal vector
*
*
*  Output:  Average normal, normalized to unit length.
*
*  Return value: none.
*/

void calc_vertex_smooth_normal(v_id,fe,norm)
vertex_id v_id;
facetedge_id fe;
REAL *norm;
{
  int i;
  REAL y[MAXCOORD],z[MAXCOORD];
  REAL fnorm[MAXCOORD];
  facetedge_id sidea;
  facet_id f_id = get_fe_facet(fe);
  struct boundary *bdry;  /* to check staying on same boundary */
  REAL size;


  /* check for string */
  if ( web.representation != SOAPFILM )
  { kb_error(5123,"Internal error: Using calc_vertex_smooth_normal() on non-soapfilm surface.\n",
     RECOVERABLE); 
  }

  if ( !valid_id(fe) )
     kb_error(2001,"Internal error: Invalid fe in calc_vertex_smooth_normal().\n",RECOVERABLE);

  bdry = get_facet_boundary(f_id);  /* original boundary */

  get_facet_normal(f_id,norm);  /* starter normal */

  /* go around one way to edge of face */
  sidea = fe;
  for(;;)
  { edge_id e_id;
    /* go to next facet */
    sidea = get_prev_edge(sidea);
    if ( equal_id(sidea,get_next_facet(sidea)) ) break;  /* valence 1 */
    if ( !equal_id(get_next_facet(sidea),get_prev_facet(sidea)) ) break;
    e_id = get_fe_edge(sidea);
    if ( !equal_constr(f_id,e_id) ) break;
    if ( bdry != get_edge_boundary(e_id) ) break;
    sidea = fe_inverse(get_next_facet(sidea));
    if ( bdry != get_facet_boundary(get_fe_facet(sidea)) ) break;
    if ( equal_id(sidea,fe) ) goto snormalize; /* went all the way around */

    get_fe_side(sidea,z);
    get_fe_side(get_prev_edge(sidea),y);
    cross_prod(y,z,fnorm);
    for ( i = 0 ; i < SDIM ; i++ ) norm[i] += fnorm[i];

  }

  /* go around the other way */
  sidea = fe;
  while ( equal_id(get_next_facet(sidea),get_prev_facet(sidea)) )
  { edge_id e_id;
    if ( equal_id(sidea,get_next_facet(sidea)) ) break;  /* edge */
    e_id = get_fe_edge(sidea);
    if ( !equal_constr(f_id,e_id) ) break;
    if ( bdry != get_edge_boundary(e_id) ) break;
    sidea = fe_inverse(get_next_facet(sidea));
    sidea = get_next_edge(sidea);
    if ( bdry != get_facet_boundary(get_fe_facet(sidea)) ) break;

    get_fe_side(sidea,z);
    get_fe_side(get_prev_edge(sidea),y);
    cross_prod(y,z,fnorm);
    for ( i = 0 ; i < SDIM ; i++ ) norm[i] += fnorm[i];
  }

  snormalize:

  size = sqrt(SDIM_dot(norm,norm));
  if ( size <= 0.0 ) return; 
  for ( i = 0 ; i < SDIM ; i++ ) norm[i] /= size;
}

/***************************************************************************
*
* function: lagrange_edge_normal()
*
* purpose: compute geometric normal space at a lagrange point of an edge.
*/
int lagrange_edge_normal ARGS((vertex_id,edge_id,REAL**));
int lagrange_edge_normal(v_id,e_id,norm)
vertex_id v_id;
edge_id e_id;
REAL **norm;
{
  MAT2D(vx,MAXVCOUNT,MAXCOORD);
  struct gauss_lag *gl = &gauss_lagrange[1][web.gauss1D_order];
  vertex_id *v = get_edge_vertices(e_id);
  int k,m,i;
  int ctrl = web.skel[EDGE].ctrlpts;
  REAL tang[MAXCOORD],*t=tang;

  get_edge_verts(e_id,vx,NULL);
  for ( k = 0 ; k < ctrl ; k++ ) if ( v[k] == v_id ) break;
#ifdef VOLUMEGRAD
  /* volume gradient */
  norm[0] = norm[1] = 0.0;
  for ( m = 0 ; m < gl->gnumpts ; m++ )
  { REAL y,dx;
    for ( y = 0.0, dx = 0.0, j = 0 ; j < ctrl ; j++ )
    { y += gl->gpoly[m][j]*vx[j][1];
      dx += gl->gpolypart[m][0][j]*vx[j][0];
    }
    norm[0] -= gl->gausswt[m]*y*gl->gpolypart[m][0][k];
    norm[1] -= gl->gausswt[m]*dx*gl->gpoly[m][k];
  }
#else
    /* geometric normal */
    for ( i = 0 ; i < SDIM ; i++ ) tang[i] = 0.0;
    for ( m = 0 ; m < ctrl ; m++ )
    { REAL prime;
      prime = gl->lpolypart[k][0][m]; 
      for ( i = 0 ; i < SDIM ; i++ )
     tang[i] += prime*vx[m][i];
    }
    return kernel_basis_rows(&t,norm,1,SDIM);
#endif
}

/***************************************************************************
*
* function: lagrange_facet_normal()
*
* purpose: compute geometric normal at a lagrange point of a facet.
*/
int lagrange_facet_normal ARGS((vertex_id,facet_id,REAL**));
int lagrange_facet_normal(v_id,f_id, norm)
vertex_id v_id;
facet_id f_id;
REAL **norm;
{ 
  vertex_id *v = get_facet_vertices(f_id);
  struct gauss_lag *gl = &gauss_lagrange[web.dimension][web.gauss2D_order];
  MAT2D(sides,MAXCOORD,MAXCOORD);
  MAT2D(vx,MAXVCOUNT,MAXCOORD);
  int ctrl = web.skel[FACET].ctrlpts;
  int i,j,k,kk;

  get_facet_verts(f_id,vx,NULL);
  for ( k = 0 ; k < ctrl ; k++ ) if ( v[k] == v_id ) break;
  if ( k >= ctrl )
     { sprintf(msg,"Internal error: Can't find vertex %d in facet %d.\n",ordinal(v_id)+1,
        ordinal(f_id)+1);
       kb_error(1900,msg,RECOVERABLE); 
     }

  for ( j = 0 ; j < SDIM ; j++ ) sides[0][j] = 0.0;
  for ( i = 1 ; i <= web.dimension ; i++ )
    for ( j = 0 ; j < SDIM ; j++ )
      for ( kk = 0, sides[i][j] = 0.0 ; kk < ctrl ; kk++ )
         sides[i][j] += gl->lpolypart[k][i-1][kk]*vx[kk][j];
  if ( inverted(f_id) ) 
     for ( j = 0 ; j < SDIM ; j++ ) sides[1][j] = -sides[1][j];
  return kernel_basis_rows(sides+1,norm,web.dimension,SDIM);
}

/***************************************************************************
*
* function: quadratic_facet_normal()
*
* purpose: compute geometric normal at a lagrange point of a facet.
*/
int qlconvert[6] = { 0, 1, 2, 4, 5, 3}; /* index conversion */
int lqconvert[6] = { 0, 1, 2, 5, 3, 4}; /* index conversion */
int quadratic_facet_normal ARGS((vertex_id,facet_id,REAL**));
int quadratic_facet_normal(v_id,f_id, norm)
vertex_id v_id;
facet_id f_id;
REAL **norm;
{
  vertex_id v[6];
  struct gauss_lag *gl = &gauss_lagrange[web.dimension][web.gauss2D_order];
  MAT2D(sides,MAXCOORD,MAXCOORD);
  MAT2D(vx,MAXVCOUNT,MAXCOORD);
  int ctrl = web.skel[FACET].ctrlpts;
  int i,j,k,kk;
  facetedge_id fe_id;

  fe_id = get_facet_fe(f_id);
  for ( i = 0 ; i < FACET_EDGES ; i++ )
     { v[qlconvert[2*i]] = get_fe_tailv(fe_id);
       v[qlconvert[2*i+1]] = get_fe_midv(fe_id);
       fe_id = get_next_edge(fe_id);
     }
  get_facet_verts(f_id,vx,NULL);
  for ( k = 0 ; k < ctrl ; k++ ) if ( v[k] == v_id ) break;
  if ( k >= ctrl )
     { sprintf(msg,"Internal error: Can't find vertex %d in facet %d.\n",ordinal(v_id)+1,
        ordinal(f_id)+1);
       kb_error(1900,msg,RECOVERABLE); 
     }

  for ( j = 0 ; j < SDIM ; j++ ) sides[0][j] = 0.0;
  for ( i = 1 ; i <= web.dimension ; i++ )
    for ( j = 0 ; j < SDIM ; j++ )
      for ( kk = 0, sides[i][j] = 0.0 ; kk < ctrl ; kk++ )
        sides[i][j] += gl->lpolypart[k][i-1][kk]*vx[lqconvert[kk]][j];
  return kernel_basis_rows(sides+1,norm,web.dimension,SDIM);
}

/***********************************************************
*
*  Function: new_calc_vertex_normal
*
*  Purpose: Calculate subspace normal to surface for purposes
*          of Hessian normal or normal motion.  Normal space
*          is essentially that spanned by the volume normals
*          of all adjacent bodies.  Idea is to rule out 
*          purely tangential motions.
*
*  Input:    vertex id
*          norm -  pointer to place for normal basis
*
*
*  Output:  Basis for normal space if not full dimension
*  Return value: Dimension of normal
*/

int new_calc_vertex_normal(v_id,norm)
vertex_id v_id;
REAL **norm; /* returned basis */
{
  int i,j,kk;
  REAL y[MAXCOORD],z[MAXCOORD];
  MAT2D(fnorm,MAXCOORD,MAXCOORD);
  MAT2D(a,2,MAXCOORD);
  facetedge_id sidea,sideb,ffe;
  struct boundary *bdry;  /* to check staying on same boundary */
  REAL size;
  int retval = 0;
  facetedge_id fe;

  if (web.representation == SIMPLEX ) return simplex_vertex_normal(v_id,norm);

  if ( hessian_special_normal_flag )
  { /* this for wet cube */
     REAL *x = get_coord(v_id); 
  /*   norm[0][0] = 1.; norm[0][1] = 1.; norm[0][2] = -2.;     */
     norm[0][0] = x[0]; norm[0][1] = x[1]; norm[0][2] = -2*x[2]; 
     return 1;
  }                             

  if ( hessian_normal_one_flag )
  { MAT2D(mat,MAXCOORD,MAXCOORD);
    facet_id f_id,startf;
    edge_id e_id, starte;

    for ( i = 0 ; i < SDIM ; i++ )
      for ( j = 0 ; j < SDIM ; j++ )
         mat[i][j] = 0.0;
    if ( web.representation == SOAPFILM )
    { 
      if ( vfacet_timestamp < top_timestamp ) make_vfacet_lists();
      f_id = startf = get_vertex_facet(v_id);
      do 
      { get_facet_normal(f_id,fnorm[0]);
        size = SDIM_dot(fnorm[0],fnorm[0]);
        for ( i = 0 ; i < SDIM ; i++ )
           for ( j = 0 ; j < SDIM ; j++ )
              mat[i][j] += fnorm[0][i]*fnorm[0][j]/size;
        f_id = get_next_vertex_facet(v_id,f_id);
      } while ( !equal_element(f_id,startf) );
    }
    else if ( web.representation == STRING )
    { 
      if ( vedge_timestamp < top_timestamp ) make_vedge_lists();
      e_id = starte = get_vertex_edge(v_id);
      do 
      { get_edge_side(e_id,fnorm[0]);
        mat[0][0] += fnorm[0][1]*fnorm[0][1];
        mat[0][1] -= fnorm[0][1]*fnorm[0][0];
        mat[1][0] -= fnorm[0][0]*fnorm[0][1];
        mat[1][1] += fnorm[0][0]*fnorm[0][0];
        e_id = get_next_tail_edge(e_id);
      } while ( !equal_element(e_id,starte) );
    }

    /* find eigenvector */
    for ( kk = 0 ; kk < 10 ; kk++ )
    { matvec_mul(mat,fnorm[0],y,SDIM,SDIM);
      matvec_mul(mat,y,fnorm[0],SDIM,SDIM);
    }

    size = sqrt(SDIM_dot(fnorm[0],fnorm[0]));
    for ( i = 0 ; i < SDIM ; i++ ) norm[0][i] = fnorm[0][i]/size;
    return 1;
  } /* end hessian_normal_one_flag */

  if ( (get_vattr(v_id) & Q_MIDEDGE) && (web.representation == STRING) )
  { 
     edge_id e_id = get_vertex_edge(v_id);
     retval = lagrange_edge_normal(v_id,e_id,norm);
     goto ncn_exit;
  }
  if ( get_vattr(v_id) & Q_MIDFACET )
  { /* Lagrange model */
     facet_id f_id = get_vertex_edge(v_id);
     retval = lagrange_facet_normal(v_id,f_id,norm);
     goto ncn_exit;
  }

  fe = get_vertex_fe(v_id);
  if ( !valid_id(fe) ) return SDIM;

  /* check for string */
  if ( web.representation == STRING )
  { edge_id e_id,ee_id,eee_id;
     e_id = get_fe_edge(fe);
     ee_id = get_next_tail_edge(e_id);
     eee_id = get_next_tail_edge(ee_id);

     if ( !equal_id(e_id,eee_id)  )
       return(SDIM);  /* triple vertex at least */

     if ( get_vattr(v_id) & Q_MIDPOINT )
     { REAL *t=y;
       if ( web.representation == STRING ) 
       { get_fe_side(fe,y);  /* geometric tangent at midpoint of quadratic */
         retval = kernel_basis_rows(&t,norm,1,SDIM);
         goto ncn_exit;
       }
     }

     if ( (web.modeltype == LAGRANGE) && (web.lagrange_order >= 2) )
       retval = lagrange_edge_normal(v_id,e_id,norm);
     else if ( web.modeltype == QUADRATIC )
     { MAT2D(vx,3,MAXCOORD);
       REAL *t = y;
       get_edge_verts(e_id,vx,NULL);
       for ( i = 0 ; i < SDIM ; i++ ) 
         y[i] = -1.5*vx[0][i] + 2*vx[1][i] - 0.5*vx[2][i];
       retval = kernel_basis_rows(&t,norm,1,SDIM);
     }
     else 
     { REAL *t = y;
       get_edge_side(e_id,y);
       if ( !equal_id(e_id,ee_id ) )
       { get_edge_side(ee_id,z);
         for ( i = 0 ; i < SDIM ; i++ ) y[i] -= z[i];
       }
       retval = kernel_basis_rows(&t,norm,1,SDIM);
     }
     goto ncn_exit;    
  }

  /* now have 2D facet */
  bdry = get_facet_boundary(get_fe_facet(fe));  /* original boundary */

  sidea = fe;
  if ( get_vattr(v_id) & Q_MIDPOINT )
  { /* quadratic model */
     MAT2D(norm2,MAXCOORD,MAXCOORD);
     facet_id f_id = get_fe_facet(sidea);
     if ( inverted(f_id) )
        sidea = inverse_id(sidea); 
     if ( !equal_id(get_next_facet(sidea),get_prev_facet(sidea)) )
       goto tripletest;
     retval = quadratic_facet_normal(v_id,get_fe_facet(sidea),norm);
     sideb = get_next_facet(sidea); 
     if ( equal_id(sidea,sideb) ) goto ncn_exit;
     sideb = inverse_id(sideb);
     retval = quadratic_facet_normal(v_id,get_fe_facet(sideb),norm2);
     for ( i = 0 ; i < retval ; i++ )
        for ( j = 0 ; j < SDIM ; j++ )
           norm[i][j] += norm2[i][j];
     goto ncn_exit;
  }
  if ( get_vattr(v_id) & Q_MIDEDGE )
  { /* Lagrange model */
     MAT2D(norm2,MAXCOORD,MAXCOORD);
     facet_id f_id = get_fe_facet(sidea);
     if ( inverted(f_id) )
        sidea = inverse_id(sidea); 
     if ( !equal_id(get_next_facet(sidea),get_prev_facet(sidea)) )
       goto tripletest;
     retval = lagrange_facet_normal(v_id,get_fe_facet(sidea),norm);
     sideb = get_next_facet(sidea); 
     if ( equal_id(sidea,sideb) ) goto ncn_exit;
     sideb = inverse_id(sideb);
     retval = lagrange_facet_normal(v_id,get_fe_facet(sideb),norm2);
     for ( i = 0 ; i < retval ; i++ )
         for ( j = 0 ; j < SDIM ; j++ )
             norm[i][j] += norm2[i][j];
     goto ncn_exit;
  }
  /* normal case of facet in N dim */
  { for ( j = 0 ; j < SDIM ; j++ )
       for ( i = 0 ; i < SDIM ; i++ ) norm[j][i] = 0.0;

     /* go around one way to edge of face */
    do
     { 
       if ( get_vattr(v_id) & AXIAL_POINT )
       { get_fe_side(sidea,z);
         get_fe_side(get_next_edge(sidea),y);
         cross_prod(y,z,fnorm[1]);
         retval = 1;
       }
       else
       { 
         switch ( web.modeltype )
         { case LINEAR: 
              get_fe_side(sidea,a[0]);
              get_fe_side(inverse_id(get_prev_edge(sidea)),a[1]);
              retval = kernel_basis_rows(a,fnorm,2,SDIM);
              break;
           case QUADRATIC: 
             retval = quadratic_facet_normal(v_id,get_fe_facet(sidea),fnorm);
             break;
           case LAGRANGE: 
             retval = lagrange_facet_normal(v_id,get_fe_facet(sidea),fnorm);  
             break;
         }
       }
       for ( j = 0 ; j < retval ; j++ )
        for ( i = 0 ; i < SDIM ; i++ ) 
         norm[j][i] += fnorm[j][i];

       /* go to next facet */
       sidea = get_prev_edge(sidea);
       if ( equal_id(sidea,get_next_facet(sidea)) ) break;  /* edge */
       sidea = fe_inverse(get_next_facet(sidea));
       if ( bdry != get_facet_boundary(get_fe_facet(sidea)) ) break;
       if ( equal_id(sidea,fe) )
       { if ( inverted(get_fe_facet(sidea)) )
            for ( i = 0 ; i < SDIM ; i++ ) norm[0][i] = -norm[0][i];
         goto ncn_exit; /* went all the way around */
       }
     }
     while ( equal_id(get_next_facet(sidea),get_prev_facet(sidea)) );
  }

tripletest:
  if ( !equal_id(get_next_facet(sidea),get_prev_facet(sidea)) )
  { /* triple line */
     REAL *t = fnorm[0];
     int triple_count = 0;
     facetedge_id start_e;
     edge_id e_id;

     if ( vedge_timestamp < top_timestamp ) make_vedge_lists(); 
     for ( i = 0 ; i < SDIM ; i++ ) fnorm[0][i] = 0.;
     start_e = e_id = get_fe_edge(sidea);
     if ( (web.modeltype==LAGRANGE) && (web.dimension >= 2) )
       retval = lagrange_edge_normal(v_id,e_id,norm);
     else
     { do 
       { e_id = get_next_tail_edge(e_id);
         ffe = get_edge_fe(e_id);
         if ( !equal_id(get_next_facet(ffe),get_prev_facet(ffe)) )
         { /* triple again */
            triple_count++;
            get_fe_side(ffe,z);
            switch ( triple_count )
            { case 1:
                for ( i = 0 ; i < SDIM ; i++ ) fnorm[0][i] += z[i]; break;
              case 2:
                for ( i = 0 ; i < SDIM ; i++ ) fnorm[0][i] -= z[i]; break;
              default: return SDIM;  /* more than 2 */
            }
         }
       } while ( !equal_element(start_e,e_id) );
       retval = kernel_basis_rows(&t,norm,1,SDIM);
     }
     goto ncn_exit;
  }
  /* go around the other way */
  sidea = fe;
  while ( equal_id(get_next_facet(sidea),get_prev_facet(sidea)) )
  {
    if ( equal_id(sidea,get_next_facet(sidea)) ) break;  /* edge */
    sidea = fe_inverse(get_next_facet(sidea));
    sidea = get_next_edge(sidea);
    if ( bdry != get_facet_boundary(get_fe_facet(sidea)) ) break;

    get_fe_side(sidea,z);
    if ( get_vattr(v_id) & AXIAL_POINT )
    { REAL *u = fnorm[0];       
      switch ( web.modeltype )
      { case LINEAR: get_fe_side(get_next_edge(sidea),y);
                       cross_prod(y,z,fnorm[0]); break;
        case QUADRATIC: 
          quadratic_facet_normal(v_id,get_fe_facet(sidea),&u);
          break;
        case LAGRANGE: 
          lagrange_facet_normal(v_id,get_fe_facet(sidea),&u);  
          break;
      }
    }
    else
    { 
         switch ( web.modeltype )
         { case LINEAR: 
              get_fe_side(sidea,a[0]);
              get_fe_side(inverse_id(get_prev_edge(sidea)),a[1]);
              retval = kernel_basis_rows(a,fnorm,2,SDIM);
              break;
           case QUADRATIC: 
             retval = quadratic_facet_normal(v_id,get_fe_facet(sidea),fnorm);
             break;
           case LAGRANGE: 
             retval = lagrange_facet_normal(v_id,get_fe_facet(sidea),fnorm);  
             break;
         }
    }
    for ( j = 0 ; j < retval ; j++ )
      for ( i = 0 ; i < SDIM ; i++ ) 
        norm[j][i] += fnorm[j][i];
  }

  if ( !equal_id(get_next_facet(sidea),get_prev_facet(sidea)) )
     goto tripletest;
      if( inverted(get_fe_facet(sidea)) )
         for ( i = 0 ; i < SDIM ; i++ ) norm[0][i] = -norm[0][i];

ncn_exit:

  return retval;
}


/***************************************************************
*
*  Function: get_facet_normal()
*
*  Purpose: find facet normal (length = area).
*          3D only
*
*/

void get_facet_normal(f_id,normal)
facet_id f_id;
REAL *normal;
{ facetedge_id fe;
  REAL side1[MAXCOORD],side2[MAXCOORD];
  int i;

  if ( web.representation == SIMPLEX )
     kb_error(1368,"Internal error: Trying to take normal in simplex model.\n",RECOVERABLE);

  if ( web.modeltype == LINEAR )  /* better for strange symmetries */
  { MAT2D(x,FACET_VERTS,MAXCOORD);
    get_facet_verts(f_id,x,NULL);  /* in tail order */
    for ( i = 0 ; i < SDIM ; i++ )
     { side1[i] = x[1][i] - x[0][i];
       side2[i] = x[2][i] - x[1][i];
     }
  }
  else
  { fe = get_facet_fe(f_id);
    get_edge_side(get_fe_edge(fe),side1);
    fe = get_next_edge(fe);
    get_edge_side(get_fe_edge(fe),side2);
  }
  cross_prod(side1,side2,normal);
  for ( i = 0 ; i < SDIM ; i++ )
     normal[i] /= 2;  /* triangle factor */
}
  
#ifdef NEWFACETNORMAL
attempt to do normal at arbitrary point in facet.  Incomplete
/***************************************************************
*
*  Function: new_get_facet_normal()
*
*  Purpose: find facet normal (length = area).
*          Any dimension, linear, quadratic, or Lagrange model.
*          Needs barycentric coords for quadratic or Lagrange.
*
*/

void new_get_facet_normal(f_id,normal,b)
facet_id f_id;
REAL *normal;
REAL *b; /* barycentric coords */
{ facetedge_id fe;
  MAT2D(x,MAXCOORD+1,MAXCOORD);
  int i;
  int m;
  MAT2D(sides,MAXCOORD,MAXCOORD);

  normal[i] /= 2;  /* triangle factor */

  if ( web.modeltype == LINEAR )
  {
    get_facet_verts(f_id,x,NULL);  /* in tail order */

    /* fan of sides from v0 */
    for ( i = 0 ; i < web.dimension ; i++ )
    for ( j = 0 ; j < SDIM ; j++ )
          sides[i][j] = x[i+1][j] - x[0][j];

    kernel_basis_rows(sides,&normal,web.dimension,SDIM);
  }
  else if ( web.modeltype == QUADRATIC )
  {
     get_facet_verts(f_id,x,NULL);  /* in tail order */
     /* tangents at point */
     mat_mult(gpolypartial[m],x,sides,web.dimension,FACET_CTRL,SDIM);
     kernel_basis_rows(sides,&normal,web.dimension,SDIM);
  }
  else if ( web.modeltype == LAGRANGE )
  {
     int ctrl = web.skel[FACET].ctrlpts;
     int dim = web.dimension;
     struct gauss_lag *gl = &gauss_lagrange[dim][web.gauss2D_order];
     get_facet_verts(f_id,x,NULL); 
     /* tangent vectors at gauss points */
     mat_mult(gl->gpolypart[m],x,sides,dim,ctrl,SDIM);
     kernel_basis_rows(sides,&normal,web.dimension,SDIM);
  }

}
#endif  
  
/***************************************************************
*
*  Function: get_edge_verts()
*
*  Purpose: to get all vertices of an edge, nicely 
*          displaced relative to first vertex of edge in case of symmetry gp
*          Quadratic model: midv in verts[1]
*
*/

void get_edge_verts(e_id,verts,wraps)
edge_id e_id;
REAL **verts;    /* third spot NULL if don't want midpts in quadratic */
WRAPTYPE *wraps;  /* wraps of vertices; NULL if not wanted */
{
  int j,k;
  WRAPTYPE wrap = 0;
  REAL *x;

  if ( web.representation == SIMPLEX )
     kb_error(1369,"Internal error: Can't do get_edge_verts for simplex model.\n",RECOVERABLE);

  if ( web.modeltype == LAGRANGE )
  { vertex_id *v = get_edge_vertices(e_id);
     int n = web.skel[EDGE].ctrlpts;
     if ( web.symmetry_flag )
     {
      wrap = get_edge_wrap(e_id);
     }
     for ( j = 0 ; j < n ; j++ )
     { x = get_coord(v[j]);
       if ( wrap && (j > 0) && (inverted(e_id) || (j==n-1)))
         (*sym_wrap)(x,verts[j],wrap);
       else
       for ( k = 0 ; k < SDIM ; k++ )
         verts[j][k] = x[k];
     }
     return;
  }

  if ( web.symmetry_flag )
    {
      x = get_coord(get_edge_tailv(e_id));
      for ( j = 0 ; j < SDIM ; j++ ) verts[0][j] = x[j];
      wrap = get_edge_wrap(e_id);
      if ( (web.modeltype == QUADRATIC) && verts[2] ) 
      { if ( wraps ){wraps[0] = 0;  wraps[2] = wrap; }
         x = get_coord(get_edge_headv(e_id));
         (*sym_wrap)(x,verts[2],wrap);
         x = get_coord(get_edge_midv(e_id));
         if ( inverted(e_id) )
         { if ( wraps ) wraps[1] = wrap;
           (*sym_wrap)(x,verts[1],wrap);
         }
         else
         { for ( j = 0 ; j < SDIM ; j++ ) verts[1][j] = x[j];
           if ( wraps ) wraps[1] = 0;
         }
      } 
      else
      { if ( wraps ){wraps[0] = 0;  wraps[1] = wrap; }
         x = get_coord(get_edge_headv(e_id));
         (*sym_wrap)(x,verts[1],wrap);
      }
    }
  else
    { x = get_coord(get_edge_tailv(e_id));
      for ( j = 0 ; j < SDIM ; j++ ) verts[0][j] = x[j];
      x = get_coord(get_edge_headv(e_id));
      if ( (web.modeltype == QUADRATIC) && verts[2] ) 
      { for ( j = 0 ; j < SDIM ; j++ ) verts[2][j] = x[j];
        x = get_coord(get_edge_midv(e_id));
        for ( j = 0 ; j < SDIM ; j++ ) verts[1][j] = x[j];
      }
      else
       for ( j = 0 ; j < SDIM ; j++ ) verts[1][j] = x[j];
    }
}
  
/***************************************************************
*
*  Function: get_facet_verts()
*
*  Purpose: to get all three vertices of a facet, nicely 
*          displaced relative to first vertex of facet in case of symmetry gp
*
*/

void get_facet_verts(f_id,verts,wraps)
facet_id f_id;
REAL **verts;    /* fourth spot NULL if don't want midpts in quadratic */
WRAPTYPE *wraps;  /* wraps of vertices; NULL if not wanted */
{
  facetedge_id fe;
  int i,j;
  WRAPTYPE wrap,fewrap;
  REAL *x;

  if ( web.modeltype == LAGRANGE )
    { vertex_id *v = get_facet_vertices(f_id);
      int ctrlpts = web.skel[FACET].ctrlpts;
      int n = web.lagrange_order;

      f_id = positive_id(f_id); /* since vertices in particular order */

      for ( i = 0 ; i < ctrlpts ; i++ )
      { x = get_coord(v[i]);
        for ( j = 0 ; j < SDIM ; j++ )
          verts[i][j] = x[j];
      }
      if ( web.symmetry_flag )
      {
        if ( wraps ) 
          for ( j = 0 ; j < ctrlpts ; j++ ) wraps[j] = 0;
    
        /* gotta go through the 3 edges */
        /* first edge */
        fe = get_facet_fe(f_id);
        wrap = get_fe_wrap(fe);
        if ( wrap && inverted(get_fe_edge(fe)) )
          for ( j = 1 ; j < n ; j++ )
          { (*sym_wrap)(get_coord(v[j]),verts[j],wrap);
             if ( wraps ) wraps[j] = wrap;
          }
        if ( wrap )
        { (*sym_wrap)(get_coord(v[n]),verts[n],wrap);
          if ( wraps ) wraps[n] = wrap ;
        }

        /* second edge */
        fe = get_next_edge(fe);
        if ( wrap && !inverted(get_fe_edge(fe)) )
          for ( j = 1 ; j < n ; j++ )
          { int m = n*(j+1) - j*(j-1)/2;
             (*sym_wrap)(get_coord(v[m]),verts[m],wrap);
             if ( wraps ) wraps[m] = wrap;
          }
        fewrap = get_fe_wrap(fe);
        wrap = (*sym_compose)(wrap,fewrap);
        if ( wrap && inverted(get_fe_edge(fe)) )
          for ( j = 1 ; j < n ; j++ )
          { int m = n*(j+1) - j*(j-1)/2;
             (*sym_wrap)(get_coord(v[m]),verts[m],wrap);
             if ( wraps ) wraps[m] = wrap;
          }
        if ( wrap )
        { int m = n*(n+3)/2;
          (*sym_wrap)(get_coord(v[m]),verts[m],wrap);
          if ( wraps ) wraps[m] = wrap;
        }
     
        /* third edge */ 
        fe = get_next_edge(fe);
        if ( wrap && !inverted(get_fe_edge(fe)) )
          for ( j = 1 ; j < n ; j++ )
          { int m = n*j + j*(3-j)/2;
             (*sym_wrap)(get_coord(v[m]),verts[m],wrap);
             if ( wraps ) wraps[m] = wrap;
          }
        fewrap = get_fe_wrap(fe);
        wrap = (*sym_compose)(wrap,fewrap);
        if ( wrap && inverted(get_fe_edge(fe)) )
          for ( j = 1 ; j < n ; j++ )
          { int m = n*j + j*(3-j)/2;
             (*sym_wrap)(get_coord(v[m]),verts[m],wrap);
             if ( wraps ) wraps[m] = wrap;
          }
      }
      return;
    }

  if ( (web.modeltype == QUADRATIC) && verts[3] ) 
  { get_facet_verts_q(f_id,verts,wraps); return; }

  if ( web.representation == SIMPLEX )
  { vertex_id *v = get_facet_vertices(f_id);
    for ( i = 0 ; i <= web.dimension ; i++ )
     { x = get_coord(v[i]);
        for ( j = 0 ; j < SDIM ; j++ )
          verts[i][j] = x[j];
     }
    return;
  }

  fe = get_facet_fe(f_id);
  if ( web.symmetry_flag )
  { wrap = 0;
    for ( i = 0 ; i < FACET_VERTS ; i++ )
    { 
      x = get_coord(get_fe_tailv(fe));
      if ( wraps ) wraps[i] = wrap;
      (*sym_wrap)(x,verts[i],wrap);
      fewrap = get_fe_wrap(fe);
      wrap = (*sym_compose)(wrap,fewrap);
      fe = get_next_edge(fe);
    }
  }
  else for ( i = 0 ; i < FACET_VERTS ; i++ )
    { x = get_coord(get_fe_tailv(fe));
      for ( j = 0 ; j < SDIM ; j++ )
        verts[i][j] = x[j];
      fe = get_next_edge(fe);
    }
}
  
/***************************************************************
*
*  Function: get_facet_verts_special()
*
*  Purpose: to get all three vertices of a facet, nicely 
*          displaced to fundamental cell in case of symmetry group.
*          Extra work to get them to come out close to origin.
*
*/

void get_facet_verts_special(f_id,verts,wraps)
facet_id f_id;
REAL **verts;    /* fourth spot NULL if don't want midpts in quadratic */
WRAPTYPE *wraps;  /* wraps of vertices; NULL if not wanted */
{
  facetedge_id fe;
  int i,j,ii;
  WRAPTYPE wrap,fewrap;
  REAL *x;

  if ( web.modeltype == LAGRANGE )
  { vertex_id *v = get_facet_vertices(f_id);
    int k = web.skel[FACET].ctrlpts;
    for ( i = 0 ; i < k ; i++ )
     { x = get_coord(v[i]);
       for ( j = 0 ; j < SDIM ; j++ )
          verts[i][j] = x[j];
     }
     return;
  }

  if ( (web.modeltype == QUADRATIC) && verts[3] ) 
  { get_facet_verts_q(f_id,verts,wraps); return; }

  if ( web.representation == SIMPLEX )
  { vertex_id *v = get_facet_vertices(f_id);
    for ( i = 0 ; i <= web.dimension ; i++ )
     { x = get_coord(v[i]);
       for ( j = 0 ; j < SDIM ; j++ )
          verts[i][j] = x[j];
     }
     return;
  }

  fe = get_facet_fe(f_id);
  if ( web.symmetry_flag )
  {
    /* start with point closest to origin */
    if ( !(sym_flags & HAS_FIXED_PTS) )
    { REAL dd,mindd;
      int mini=0;
      facetedge_id minfe=fe;
      mindd = 1e40;
      for ( i = 0 ; i < FACET_VERTS ; i++, fe = get_next_edge(fe) )
      { x = get_coord(get_fe_tailv(fe));
         dd = SDIM_dot(x,x);
         if ( dd < 0.9*mindd ) { mindd = dd; mini = i; minfe = fe;}
      }
      i = mini;
      fe = minfe;
    }
    else { i = 0; }
    wrap = 0;
    for ( ii = 0 ; ii < FACET_VERTS ; ii++ )
    { int jj = (i+ii)%FACET_VERTS;
      x = get_coord(get_fe_tailv(fe));
      if ( wraps ) wraps[jj] = wrap;
      (*sym_wrap)(x,verts[jj],wrap);
      fewrap = get_fe_wrap(fe);
      wrap = (*sym_compose)(wrap,fewrap);
      fe = get_next_edge(fe);
    }
  }
  else for ( i = 0 ; i < FACET_VERTS ; i++ )
    { x = get_coord(get_fe_tailv(fe));
      for ( j = 0 ; j < SDIM ; j++ )
        verts[i][j] = x[j];
      fe = get_next_edge(fe);
    }
}
  
/***************************************************************
*
*  Function: get_facet_verts_q()
*
*  Purpose: to get all vertices of a facet, nicely 
*          displaced to fundamental cell in case of symmetry group.
*          Quadratic model.  Midpoints interleaved with corners.
*
*/

void get_facet_verts_q(f_id,verts,wraps)
facet_id f_id;
REAL **verts;
WRAPTYPE *wraps;  /* NULL if not wanted */
{
  facetedge_id fe;
  int i,j,ii;
  WRAPTYPE wrap,fewrap;
  REAL *x;

  if ( web.representation == SIMPLEX )
    { kb_error(1370,"No quadratic model with simplices.\n",RECOVERABLE);

    }

  fe = get_facet_fe(f_id);
  if ( web.symmetry_flag )
  {
    i = 0;
    wrap = 0;
    for ( ii = 0 ; ii < FACET_VERTS ; ii++ )
    { int jj = (i+ii)%FACET_VERTS;
      x = get_coord(get_fe_tailv(fe));
      if ( wraps ) wraps[2*jj] = wrap;
      (*sym_wrap)(x,verts[2*jj],wrap);

      x = get_coord(get_fe_midv(fe));

      if ( !inverted(get_fe_edge(fe)) )
      { if ( wraps ) wraps[2*jj+1] = wrap;
         (*sym_wrap)(x,verts[2*jj+1],wrap);
      }

      fewrap = get_fe_wrap(fe);
      wrap = (*sym_compose)(wrap,fewrap);

      if ( inverted(get_fe_edge(fe)) )
      { if ( wraps ) wraps[2*jj+1] = wrap;
         (*sym_wrap)(x,verts[2*jj+1],wrap);
      }

      fe = get_next_edge(fe);
    }
  }
  else for ( i = 0 ; i < FACET_VERTS ; i++ )
    { x = get_coord(get_fe_tailv(fe));
      for ( j = 0 ; j < SDIM ; j++ )
        verts[2*i][j] = x[j];
      x = get_coord(get_fe_midv(fe));
      for ( j = 0 ; j < SDIM ; j++ )
        verts[2*i+1][j] = x[j];
      fe = get_next_edge(fe);
    }
}

/*****************************************************************
*
* Function: path_open()
*
*  Purpose: Open file for reading on EVOLVERPATH
*/

#ifdef __WIN32__
FILE *path_open(char *name)    // else TC++ has conniptions
#else
FILE *path_open(name)
char *name;
#endif
{
  char *env;
  char path[200];
  int len;
  FILE *fd;

  env = getenv("EVOLVERPATH");

  /* try given name */
  strncpy(path,name,sizeof(path));
  while ( (fd = fopen(path,"r")) == NULL)
  { /* try paths in EVOLVERPATH */
    if ( env == NULL ) break;
    len = strcspn(env,ENVPATHCHAR);
    if ( len == 0 ) break;
    strncpy(path,env,len);
    path[len] = PATHCHAR;
    strncpy(path+len+1,name,sizeof(path)-len-2);
    if ( env[len] == 0 ) env = NULL; /* end of EVOLVERPATH */
    else env += len+1;
  } 
  
  /* try .fe extension */
  if ( fd == NULL)
  {
    env = getenv("EVOLVERPATH");
    strncpy(path,name,sizeof(path));
    strcat(path,".fe");
    while ( (fd = fopen(path,"r")) == NULL)
     { /* try paths in EVOLVERPATH */
        if ( env == NULL ) break;
        len = strcspn(env,ENVPATHCHAR);
        if ( len == 0 ) break;
        strncpy(path,env,len);
        path[len] = PATHCHAR;
        strncpy(path+len+1,name,sizeof(path)-len-2);
        strcat(path,".fe");
        if ( env[len] == 0 ) env = NULL; /* end of EVOLVERPATH */
        else env += len+1;
     } 
  }
  
  return fd;
}


/*******************************************************************
*
*  function: calc_periods()
*
*  purpose: calculate torus periods from expressions
*/

void calc_periods(mode)
int mode;  /* NO_ADJUST_VOLUMES or ADJUST_VOLUMES */
{ int i,j;
  REAL value;
  REAL old_torusv = web.torusv;
  int k;

  for ( i = 0 ; i < SDIM ; i++ )
  {
    for ( j = 0 ; j < SDIM ; j++ )
     {
        value = eval(&torus_period_expr[i][j],NULL,NULLID);
        web.inverse_periods[j][i] = web.torus_period[i][j] = value;
     }
  }
  web.torusv = det_adjoint(web.inverse_periods,SDIM);
  if ( web.torusv == 0.0 )
      kb_error(1377,"Degenerate torus unit cell.\n",RECOVERABLE);

  for ( i = 0 ; i < SDIM ; i++ )
    for ( j = 0 ; j < SDIM ; j++ )
     web.inverse_periods[i][j] /= web.torusv;

  /* if change in periods, adjust volumes and volconsts */
  if ( (mode == ADJUST_VOLUMES) && (old_torusv != 0.0) )
  { body_id b_id;
     value = web.torusv/old_torusv;
     FOR_ALL_BODIES(b_id)
     { set_body_volume(b_id,value*(get_body_volume(b_id)));
       save_body_volume(b_id);
       if ( get_battr(b_id) & FIXEDVOL )
          set_body_fixvol(b_id,value*(get_body_fixvol(b_id)));
       set_body_volconst(b_id,value*(get_body_volconst(b_id)));
       if ( everything_quantities_flag )
       { struct gen_quant *q = GEN_QUANTS+get_body_volquant(b_id);
         q->value *= value;
         q->target *= value;
         q->volconst *= value;
       }
     }
  }

  /* adjust view transform generators */
  if ( view_transform_gens == NULL )
  { view_transform_gens = dmatrix3(SDIM,SDIM+1,SDIM+1);
    transform_gen_count = SDIM;
  }
  if ( transform_gen_swap == NULL )
     transform_gen_swap = (int*)mycalloc(SDIM,sizeof(int));
  for ( j=transform_gen_count-SDIM, k=0 ; j<transform_gen_count ; k++,j++ )
  { matcopy(view_transform_gens[j],identmat,SDIM+1, SDIM+1);
    for ( i = 0 ; i < SDIM ; i++ )
    view_transform_gens[j][i][SDIM] = web.torus_period[k][i];
  }
  if (transform_expr && *transform_expr)
     transform_gen_expr(transform_expr);
}
  
/***************************************************
* 
*  Handy for graphics function pointers for 
*  functions that have nothing to do.
*/

void null_function() {}

/************************************************************
*
*  For uninitialized function pointers.
*
*/

void bad_function()
{ kb_error(1371,"Internal error:  Using uninitialized function pointer.\n",
     WARNING);
}

/*****************************************************************
*
* Function: distance()
*
*  Purpose: Finds distance between two vertices.
*/

REAL distance(v1,v2)
vertex_id v1,v2;
{
  REAL *c1,*c2;
  REAL sum;
  int i;

  c1 = get_coord(v1);
  c2 = get_coord(v2);
  sum = 0.0;
  for ( i = 0 ; i < SDIM ; i++ )
     sum += (c1[i] - c2[i])*(c1[i] - c2[i]);
  return sqrt(sum);
}

/***************************************************************************
*
* function: grule()
*
* purpose: Calculate abscissas and weights for gaussian quadrature
*         on [0,1].
*
* Adapted from a FORTRAN routine GRULE in a book on numerical integration 
*/

void grule(n,x,w)
int n;  /* number of points */
REAL *x;    /* for abscissas  x[0] ... x[n-1] */
REAL *w;    /* for weights     w[0] ... w[n-1] */
{
  REAL pkm1,pk,t,t1,pkp1,den,d1,dpn,d2pn,d3pn,d4pn,u,v,h,p,dp,fx,x0;
  int m,e1,i,k;
  
  if ( n < -1 )
  { /* rectangle rule */
    for ( k = 0 ; k < abs(n) ; k++ )
      { x[k] = k/(REAL)(abs(n)-1);
         w[k] = 1/(REAL)abs(n);
      }
  }
  m = (n+1)/2;
  e1 = n*(n+1);
  for ( i = 1 ; i <= m ; i++ )
  {
    /* original calculation on [-1,1] */
    t = (4*i - 1)*M_PI/(4*n+2);
    x0 = (1 - (1 - 1.0/n)/(8*n*n))*cos(t);
    pkm1 = 1.0;
    pk = x0;
    for ( k = 2 ; k <= n ; k++ )
     {
        t1 = x0*pk;
        pkp1 = t1 - pkm1 - (t1 - pkm1)/k + t1;
        pkm1 = pk;
        pk = pkp1;
     }
    den = 1 - x0*x0;
    d1 = n*(pkm1 - x0*pk);
    dpn = d1/den;
    d2pn = (2*x0*dpn - e1*pk)/den;
    d3pn = (4*x0*d2pn + (2 - e1)*dpn)/den;
    d4pn = (6*x0*d3pn + (6 - e1)*d2pn)/den;
    u = pk/dpn;
    v = d2pn/dpn;
    h = -u*(1 + 0.5*u*(v + u*(v*v - d3pn/3/dpn)));
    p = pk + h*(dpn + 0.5*h*(d2pn + h/3*(d3pn + 0.25*h*d4pn)));
    dp = dpn + h*(d2pn + .5*h*(d3pn + h*d4pn/3));
    h = h - p/dp;
    x[i-1] = x0 + h;
    fx = d1 - h*e1*((((0.2*h*d4pn + d3pn)*0.25*h + d2pn)*h/3 + dpn)*0.5*h + pk);
    w[n-i] = w[i-1] = (1 - x[i-1]*x[i-1])/fx/fx;

    /* normalization to [0,1] */
    x[i-1] = (1 + x[i-1])/2;
    x[n-i] =  1 - x[i-1];
  }
  if ( 2*m > n ) x[m-1] = 0.5;
}


/****************************************************************
*
*  function: binom_coeff()
*
*  purpose: calculate binomial coefficient (small numbers only)
*
*/

int binom_coeff(n,k)
int n,k;
{ int i, c;

  if ( n < k ) return 0;
  if ( (n-k) < k ) k = n - k; 
  for ( i = 0, c = 1 ; i < k ; i++ )
  { c *= n - i; c /= i+1; }
  return c;
}

/**********************************************************************
*
*  function: set_e_phase_density()
*
*  purpose: calculate edge density from phases of neighboring facets.
*/
void set_e_phase_density(e_id)
edge_id e_id;
{ facetedge_id fe = get_edge_fe(e_id);
  int i = get_f_phase(get_fe_facet(fe));      
  int j = get_f_phase(get_fe_facet(get_next_facet(fe)));      
      set_edge_density(e_id,phase_data[i][j]);
}

/**********************************************************************
*
*  function: set_f_phase_density()
*
*  purpose: calculate edge density from phases of neighboring bodies.
*/
void set_f_phase_density(f_id)
facet_id f_id;
{ 
  int i = get_b_phase(get_facet_body(f_id));     
  int j = get_b_phase(get_facet_body(inverse_id(f_id)));      
  set_attr(f_id,DENSITY);
  set_facet_density(f_id,phase_data[i][j]);
}

/**********************************************************************
*
*  function: make_vedge_lists()
*
*  purpose: construct from scratch edge lists aound vertices.
*    These are singly linked circular lists with links residing
*    in edge structures.
*/

void make_vedge_lists()
{ vertex_id v_id;
  edge_id e_id, ee_id,eee_id,ei_id;

  if ( web.representation == SIMPLEX ) return;

  /* start by making one-element loops on edges pointed to
      by vertex facet_edge link */
  FOR_ALL_VERTICES(v_id)
    { 
      e_id = get_vertex_edge(v_id);
      if ( !valid_id(e_id) ) continue;
      set_next_tail_edge(e_id,e_id);
    }
  /* now splice in all other edges */
  FOR_ALL_EDGES(e_id)
    { v_id = get_edge_tailv(e_id);
      ee_id = get_vertex_edge(v_id);
      if ( valid_id(ee_id) && !equal_id(e_id,ee_id) )
      { /* splice in after first one */
        eee_id = get_next_tail_edge(ee_id);
        set_next_tail_edge(ee_id,e_id);
        set_next_tail_edge(e_id,eee_id);
      }
      ei_id = inverse_id(e_id);
      v_id = get_edge_tailv(ei_id);
      ee_id = get_vertex_edge(v_id);
      if ( !valid_id(ee_id) ) continue;
      if ( !equal_id(ei_id,ee_id) )
      { /* splice in after first one */
        eee_id = get_next_tail_edge(ee_id);
        set_next_tail_edge(ee_id,ei_id);
        set_next_tail_edge(ei_id,eee_id);
      }
    }
  vedge_timestamp = top_timestamp; 
}


/**********************************************************************
*
*  function: make_vfacet_lists()
*
*  purpose: construct from scratch facet lists aound vertices.
*    These are singly linked circular lists with links residing
*    in facet structures.
*/

void make_vfacet_lists()
{ vertex_id v_id;
  facet_id f_id, ff_id,fff_id;
  int i;
  int facet_verts;

  /* set all facet vertices */
  facet_verts = web.skel[FACET].ctrlpts;
  expand_attribute(FACET,F_NEXT_VFACET_ATTR,facet_verts);
  if ( web.representation != SIMPLEX ) /* set up facet vertex lists */
  { if ( web.modeltype != LAGRANGE )
      FOR_ALL_FACETS(f_id)
      { facetedge_id fe = get_facet_fe(f_id);
        vertex_id *fv = get_facet_vertices(f_id);

        for ( i = 0 ; i < FACET_VERTS ; i++ )
        { 
           v_id = get_fe_tailv(fe);
           fv[i] = v_id;
           fe = get_next_edge(fe);
           if ( web.modeltype == QUADRATIC )
             fv[i+FACET_VERTS] = get_edge_midv(get_fe_edge(fe));
        }
      }
    }
  else FOR_ALL_FACETS(f_id) /* simplex vertex to facet links */
  { vertex_id *fv = get_facet_vertices(f_id);
    for ( i = 0 ; i < facet_verts ; i++ )
    set_vertex_edge(fv[i],f_id);
  }
  /* start by making one-element loops on facets */
  FOR_ALL_VERTICES(v_id)
  {
     f_id = get_vertex_facet(v_id);
     if ( valid_id(f_id) )
     { set_next_vertex_facet(v_id,f_id,f_id);
       set_first_vertex_facet(v_id,f_id);
     }
  }
  FOR_ALL_FACETS(f_id)
  { 
    vertex_id *fv = get_facet_vertices(f_id);
    for ( i = 0 ; i < facet_verts ; i++ )
    { v_id = fv[i]; 
      ff_id = get_vertex_facet(v_id);
      if ( !equal_id(f_id,ff_id) )
      { /* splice in after first one */
        fff_id = get_next_vertex_facet(v_id,ff_id);
        if ( !equal_element(f_id,fff_id) )
        { set_next_vertex_facet(v_id,ff_id,f_id);
          set_next_vertex_facet(v_id,f_id,fff_id);
        }
      }
    }
  }
  vfacet_timestamp = top_timestamp; /* so can call get_vertex_facet */
}

/**********************************************************************
*
*  function: make_bfacet_lists()
*
*  purpose: construct from scratch facet lists aound bodies.
*    These are singly linked circular lists with links residing
*    in facet structures.
*/

void make_bfacet_lists()
{ body_id b_id;
  facet_id f_id, ff_id,fff_id,fi_id;

  if ( web.skel[BODY].count == 0 ) return;
  expand_attribute(FACET,F_NEXT_BFACET_ATTR,2);
  /* start by making one-element loops on facets pointed to
      by body facet_edge link */
  FOR_ALL_BODIES(b_id)
  { /* make sure legal pointer */
    f_id = get_fe_facet(get_body_fe(b_id));
    if ( !valid_id(f_id) || !equal_id(b_id,get_facet_body(f_id)) )
    FOR_ALL_FACETS(f_id)
     { if ( equal_id(b_id,get_facet_body(f_id)) ) 
        {set_body_fe(b_id,get_facet_fe(f_id)); break; }
        if ( equal_id(b_id,get_facet_body(inverse_id(f_id))) ) 
        {set_body_fe(b_id,get_facet_fe(inverse_id(f_id))); break; }
     }      
     f_id = get_fe_facet(get_body_fe(b_id));
     set_next_body_facet(f_id,f_id);
  }

  /* now splice in all other facets */
  FOR_ALL_FACETS(f_id)
  { b_id = get_facet_body(f_id);
    if ( valid_id(b_id) )
    { ff_id = get_fe_facet(get_body_fe(b_id));
      if ( !equal_id(f_id,ff_id) )
      { /* splice in after first one */
        fff_id = get_next_body_facet(ff_id);
        set_next_body_facet(ff_id,f_id);
        set_next_body_facet(f_id,fff_id);
      }
    }
    fi_id = inverse_id(f_id);
    b_id = get_facet_body(fi_id);
    if ( valid_id(b_id) )
      { ff_id = get_fe_facet(get_body_fe(b_id));
        if ( !equal_id(fi_id,ff_id) )
         { /* splice in after first one */
            fff_id = get_next_body_facet(ff_id);
            set_next_body_facet(ff_id,fi_id);
            set_next_body_facet(fi_id,fff_id);
         }
      }
  }
  bfacet_timestamp = top_timestamp; 
}

/*******************************************************************
*
* function: insert_vertex_edge(v_id,e_id)
*
* purpose: fix up edge links around vertex when edge added
*
*/

void insert_vertex_edge(v_id,e_id)
vertex_id v_id;
edge_id    e_id;
{ edge_id ee_id,eee_id; 
  int n;

  if ( !equal_id(v_id,get_edge_tailv(e_id)) )
    { kb_error(1372,"Internal error: Trying to insert edge to wrong vertex.\n",WARNING);
      return;
    }
  ee_id = get_vertex_edge(v_id);
  if ( !valid_id(ee_id) )
  { set_next_tail_edge(e_id,e_id);
    set_vertex_edge(v_id,e_id);
    return;
  }
  /* go around to be sure not already there */
  eee_id = ee_id; n = 0;
  do 
  { eee_id = get_next_tail_edge(eee_id);
     if ( equal_id(eee_id,e_id) ) return;
     if ( ++n > 2*web.skel[EDGE].count+5 ) /* sanity check */
     { kb_error(1373,"Internal error: Vertex edge loop not closed.\n",WARNING);
       return;
     }
  } while ( !equal_id(eee_id,ee_id) );

  /* insert into existing list */
  ee_id = get_next_tail_edge(eee_id);
  set_next_tail_edge(eee_id,e_id);
  set_next_tail_edge(e_id,ee_id);
}

/*******************************************************************
*
* function: remove_vertex_edge(v_id,e_id)
*
* purpose: fix up edge links around vertex when edge removed
*
*/

void remove_vertex_edge(v_id,e_id)
vertex_id v_id;
edge_id    e_id;  /* tail is v_id */
{ edge_id ee_id,eee_id,eeee_id; 
  int n = 0;
  if ( !equal_id(v_id,get_edge_tailv(e_id)) )
    { kb_error(1374,"Internal error: Trying to detach edge from wrong vertex.\n",WARNING);
      return;
    }
  /* make sure edge actually in vertex edge loop */
  ee_id = get_vertex_edge(v_id);
  if ( !valid_id(ee_id)) return;
  eee_id = ee_id;
  do
  { if ( equal_id(eee_id,e_id) ) break;
    eee_id = get_next_tail_edge(eee_id);
    if ( ++n > 2*web.skel[EDGE].count ) /* sanity check */ 
    { kb_error(1375,"Internal error: Vertex edge loop not closed.\n",WARNING);
      break;
    }
  } while ( !equal_id(eee_id,ee_id) );
  if ( !equal_id(eee_id,e_id) ) return;

  ee_id = get_next_tail_edge(e_id);
  if ( equal_id(e_id,ee_id) )
    { set_vertex_edge(v_id,NULLID);
      return;
    }
  eee_id = ee_id;
  eeee_id = get_next_tail_edge(ee_id);
  n = 0;
  while ( !equal_id(e_id,eeee_id) )
    { eee_id = eeee_id;
      eeee_id = get_next_tail_edge(eee_id);
      if ( ++n > 2*web.skel[EDGE].count ) /* sanity check */
      { kb_error(1376,"Internal error: Vertex edge loop not closed.\n",WARNING);
        break;
      }
    }
  set_next_tail_edge(eee_id,ee_id);
  set_next_tail_edge(e_id,e_id);
  set_vertex_edge(v_id,ee_id);
}

/*******************************************************************
*
*  function: get_vertex_facet()
*
*  purpose: find some facet adjoining vertex.
*/

facet_id get_vertex_facet(v_id)
vertex_id v_id;
{ edge_id e_id,ee_id;
  facetedge_id fe_id;
  facet_id f_id;

  if ( web.representation == SIMPLEX ) return get_vertex_edge(v_id);
  if ( get_vattr(v_id) & Q_MIDFACET ) return get_vertex_edge(v_id);

  e_id = get_vertex_edge(v_id);
  if ( !valid_id(e_id) ) return NULLID;
  ee_id = e_id;
  do
  { fe_id = get_edge_fe(ee_id);
     if ( valid_id(fe_id) ) 
     { f_id = get_fe_facet(fe_id);
       if ( inverted(f_id) ) return inverse_id(f_id);
       return f_id;
     }
     ee_id = get_next_tail_edge(ee_id);
  } while ( ee_id != e_id );
  return  NULLID;
}
/************************************************************************
*
* function: count_fixed_vol()
*
* purpose: return number of fixed volumes and quantities
*
*/

int count_fixed_vol()
{ int retval = 0;
  body_id bi_id;  /* identifier for body i */
  int i,n;
  struct gen_quant *gq;

  if ( !everything_quantities_flag && !web.pressure_flag )
    FOR_ALL_BODIES(bi_id)
    if ( get_battr(bi_id) & (FIXEDVOL) ) retval++;
  for ( i = 0 ; i < web.quantity_count ; i++ )
    if ( get_quant(i)->attr & QFIXED ) retval++;
  for (  n = 0, gq = GEN_QUANTS ; n < gen_quant_count ; gq++,n++ )
    if ( gq->flags & Q_FIXED )
        retval++;
  return retval;
}

/*****************************************************************
*
*  Random number generator using the Mersenne Twister.
*

* A C-program for MT19937B: Integer Version 
* genrand() generate one pseudorandom integer which is  
* uniformly distributed among the 32bit unsigned integers
* sgenrand(seed) set initial values to the working area of 624 words.
* sgenrand(seed) must be called once before calling genrand()    
* (seed is any integer except 0).                    

* Modified to produce batch of drands by K. Brakke 

* Reference: 
   Makoto Matsumoto and Takuji Nishimura, 
      Mersenne Twister: A 623-Dimensionally Equidistributed Uniform
      Pseudo-Random Number Generator, ACM Transactions on Modeling and
      Computer Simulation, Vol. 8, No. 1, January 1998, pp. 3-30.
      In ACM on-line Library, 
   http://www.acm.org/pubs/citations/journals/tomacs/1998-8-1/p3-matsumoto/
*/
/* 
    LICENCE CONDITIONS: 

        Matsumoto and Nishimura consent to GNU General 
        Public Licence for this code.
*/


/* Period parameters */  
#define N  624
#define NR 312
#define M 397
#define MATRIX_A 0x9908b0df   /* constant vector a */
#define UPPER_MASK 0x80000000 /* most significant w-r bits */
#define LOWER_MASK 0x7fffffff /* least significant r bits */

/* for tempering */   
#define TEMPERING_MASK_B 0x9d2c5680
#define TEMPERING_MASK_C 0xefc60000
#define TEMPERING_SHIFT_U(y)  (y >> 11)
#define TEMPERING_SHIFT_S(y)  (y << 7)
#define TEMPERING_SHIFT_T(y)  (y << 15)
#define TEMPERING_SHIFT_L(y)  (y >> 18)

static unsigned long ptgfsr[N]; /* set initial seeds: N = 624 words */
static REAL drands[NR];  /* batch of random doubles */
static int initflag;
static REAL rfactor = 0.0;
static int k = N;
static int kr = NR;
void sgenrand(unsigned long);

void
sgenrand(seed)
unsigned long seed; /* seed should not be 0 */
{
  int k;
  
  /* setting initial seeds to ptgfsr[N] using     */
  /* the generator Line 25 of Table 1 in      */
  /* [KNUTH 1981, The Art of Computer Programming */
  /*    Vol. 2 (2nd Ed.), pp102]          */

  ptgfsr[0]= seed & 0xffffffff;
  for (k=1; k<N; k++)
    ptgfsr[k] = (69069 * ptgfsr[k-1]) & 0xffffffff;
}

void kb_initr(ind)
int ind;
{ sgenrand((unsigned long)ind); 
  rfactor = 1.0/( (REAL)0xffffffff + 1.0) ;
  initflag = 1;
  k = N; kr = NR;
}

REAL kb_drand()
{
  unsigned long y;
  static unsigned long mag01[2]={0x0, MATRIX_A};
  /* mag01[x] = x * MATRIX_A  for x=0,1 */
  
  if(kr == NR){ /* generate N words at one time, and NR drands */
    int kk;
    if ( initflag == 0 ) kb_initr(1);
    for (kk=0;kk<N-M;kk++) {
      y = (ptgfsr[kk]&UPPER_MASK)|(ptgfsr[kk+1]&LOWER_MASK);
      ptgfsr[kk] = ptgfsr[kk+M] ^ (y >> 1) ^ mag01[y & 0x1];
    }
    for (;kk<N-1;kk++) {
      y = (ptgfsr[kk]&UPPER_MASK)|(ptgfsr[kk+1]&LOWER_MASK);
      ptgfsr[kk] = ptgfsr[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1];
    }
    y = (ptgfsr[N-1]&UPPER_MASK)|(ptgfsr[0]&LOWER_MASK);
    ptgfsr[N-1] = ptgfsr[M-1] ^ (y >> 1) ^ mag01[y & 0x1];

    for ( k = 0 ; k < N ; k++ )
    { y = ptgfsr[k];
      y ^= TEMPERING_SHIFT_U(y);
      y ^= TEMPERING_SHIFT_S(y) & TEMPERING_MASK_B;
      y ^= TEMPERING_SHIFT_T(y) & TEMPERING_MASK_C;
      y &= 0xffffffff; /* you may delete this line if word size = 32 */
      y ^= TEMPERING_SHIFT_L(y);
      ptgfsr[k] = y;
    }

    for ( kr = 0 ; kr < NR ; kr++ )
      drands[kr] = (ptgfsr[2*kr]*rfactor + ptgfsr[2*kr+1])*rfactor;
    
    k = 0;
    kr = 0;
  }
  
  return drands[kr++];
}

